import { swel } from "../deps/swel.ts";
import { Callback } from "../utils.ts";

export class TextInputWithValidation {
	private input = swel("input");
	private errorText = swel("div", { className: "error-message" }, "&nbsp;");
	public el = swel("div", { className: "validated-input" }, [
		this.input,
		this.errorText,
	]);

	set placeholder(s: string) {
		this.input.placeholder = s;
	}

	set onInput(cb: Callback<[string]>) {
		this.input.oninput = () => cb(this.input.value);
	}

	set error(msg: string | undefined) {
		if (msg) {
			this.el.classList.add("error");
			this.errorText.replaceChildren(msg);
		} else {
			this.el.classList.remove("error");
			this.errorText.replaceChildren("&nbsp;");
		}
	}

	set value(v: string) {
		this.input.value = v;
	}
}

export class LabeledCheckbox {
	private checkbox = swel("input", { type: "checkbox" });
	private labelText: ReturnType<typeof document.createTextNode>;
	private label: HTMLLabelElement;

	constructor(p: { labelText: string; checked?: boolean }) {
		this.checkbox.checked = p.checked ?? false;
		this.labelText = document.createTextNode(p.labelText);
		this.label = swel("label", [this.labelText, this.checkbox]);
	}

	get el() {
		return this.label;
	}

	set checked(c: boolean) {
		this.checkbox.checked = c;
	}

	set onChange(cb: Callback<[boolean]>) {
		this.checkbox.onchange = () => cb(this.checkbox.checked);
	}
}

export type Choice<V> = {
	value: V;
	display: string;
};

export class LabeledChoices<V> {
	private label: HTMLLabelElement;
	private select: HTMLSelectElement;
	private _choices: Choice<V>[] = [];

	constructor(labelText: string) {
		this.select = swel("select");
		this.label = swel("label", [labelText, this.select]);
	}

	set choices(choices: typeof this._choices) {
		this._choices = choices;
		this.select.replaceChildren(
			...this._choices.map((c, i) =>
				swel("option", { value: `${i}` }, c.display),
			),
		);
	}

	get el() {
		return this.label;
	}

	set onChange(cb: Callback<[V]>) {
		this.select.onchange = () => {
			cb(this._choices[Number(this.select.value)].value);
		};
	}
}
