import {
	Field,
	FieldKind,
	FieldKinds,
	FieldOptions,
	MaybePersistedSchema,
	ValidateSchemaResult,
	defaultOptionsForFieldKind,
	validateSchema,
} from "../../common/schema.ts";
import { swel } from "../deps/swel.ts";
import { Callback, Update } from "../utils.ts";
import {
	LabeledChoices,
	LabeledCheckbox,
	TextInputWithValidation,
} from "./input.ts";
import { RecyclerList } from "./recycler_list.ts";

class StringOptionsEditor {
	private displayAs = new LabeledChoices<FieldOptions["string"]["displayAs"]>(
		"Display as",
	);
	private container = swel("div", { className: "string-options" }, [
		this.displayAs.el,
	]);

	private _updateOptions: Update<FieldOptions["string"]> | null = null;

	constructor() {
		this.displayAs.choices = [
			{
				value: "input",
				display: "input",
			},
			{
				value: "textarea",
				display: "textarea",
			},
		];

		this.displayAs.onChange = (s) =>
			this._updateOptions?.((o) => (o.displayAs = s));
	}

	get el() {
		return this.container;
	}

	set updateOptions(u: typeof this._updateOptions) {
		this._updateOptions = u;
	}

	optionsChanged(options: FieldOptions["string"]) {}
}
class NumberOptionsEditor {
	private container = swel("div", { className: "number-options" });

	private _updateOptions: Update<FieldOptions["string"]> | null = null;

	get el() {
		return this.container;
	}

	set updateOptions(u: typeof this._updateOptions) {
		this._updateOptions = u;
	}

	optionsChanged(options: FieldOptions["number"]) {}
}

class FieldOptionsEditor {
	private fieldKindToView = {
		string: new StringOptionsEditor(),
		number: new NumberOptionsEditor(),
	};

	private container = swel("div", { className: "options-editor" });

	private _updateField: Update<Field> | null = null;
	set updateField(update: Update<Field>) {
		this._updateField = update;
	}

	get el() {
		return this.container;
	}

	constructor() {
		Object.values(this.fieldKindToView).forEach((v) => {
			v.updateOptions = (cb) =>
				this._updateField?.((v) => cb(v.options as any));
		});
	}

	fieldChanged(field: Field) {
		const view = this.fieldKindToView[field.kind];
		view.optionsChanged(field.options as any);

		this.container.replaceChildren(view.el);
	}
}

class FieldEditor {
	private nameInput = new TextInputWithValidation();
	private kind = new LabeledChoices<FieldKind>("Kind");
	private required = new LabeledCheckbox({ labelText: "Required" });
	private optionsEditor = new FieldOptionsEditor();

	private container = swel("div", { className: "field" }, [
		this.nameInput.el,
		this.kind.el,
		this.optionsEditor.el,
		this.required.el,
	]);
	private _updateField: Update<Field> | null = null;
	set updateField(update: Update<Field>) {
		this._updateField = update;
	}

	constructor() {
		this.nameInput.placeholder = "Field name";

		this.kind.choices = FieldKinds.map((k) => ({
			display: k,
			value: k,
		}));

		this.nameInput.onInput = (s) => this._updateField?.((f) => (f.name = s));
		this.required.onChange = (s) =>
			this._updateField?.((f) => (f.required = s));
		this.kind.onChange = (k) =>
			this._updateField?.((f) => {
				f.kind = k;
				f.options = defaultOptionsForFieldKind(k);
			});

		this.optionsEditor.updateField = (cb) => this._updateField?.((f) => cb(f));
	}

	get el() {
		return this.container;
	}

	set checked(c: boolean) {
		this.required.checked = c;
	}

	fieldChanged(field: Field, error: string | undefined) {
		this.nameInput.value = field.name;
		this.nameInput.error = error;

		this.optionsEditor.fieldChanged(field);
	}
}

class FieldEditors {
	private _updateField: ((index: number) => Update<Field>) | null = null;

	private list = new RecyclerList(
		() => new FieldEditor(),
		(view) => view.el,
		(view, model: [Field, string | undefined], index) => {
			const [field, error] = model;

			view.fieldChanged(field, error);
			view.updateField = (field) => this._updateField?.(index)(field);
		},
	);

	private addFieldButton = swel(
		"button",
		{
			className: "add",
		},
		"Add field",
	);
	private container = swel("div", { className: "fields" }, [
		this.list.el,
		this.addFieldButton,
	]);

	get el() {
		return this.container;
	}

	set updateField(fn: Exclude<typeof this._updateField, null>) {
		this._updateField = fn;
	}

	set onAdd(cb: Callback<[Field]>) {
		this.addFieldButton.onclick = () =>
			cb({
				kind: "string",
				options: defaultOptionsForFieldKind("string"),
				name: "",
				required: false,
			});
	}

	fieldsChanged(fields: Field[], errors: ValidateSchemaResult) {
		this.list.changed(
			fields.map((f, i) => [f, errors[`fields.${i}.name`]?.[0]]),
		);
	}
}

class Editor {
	private nameInput = new TextInputWithValidation();
	private fields = new FieldEditors();
	private container = swel("div", { className: "schema-editor" }, [
		this.nameInput.el,
		this.fields.el,
	]);
	private _updateSchema: Update<MaybePersistedSchema> | null = null;

	constructor() {
		this.nameInput.placeholder = "Schema name";

		this.nameInput.onInput = (s) => {
			this._updateSchema?.((sc) => (sc.name = s));
		};

		this.fields.onAdd = (f) => {
			this._updateSchema?.((sc) => sc.fields.push(f));
		};

		this.fields.updateField = (idx) => (cb) => {
			this._updateSchema?.((sc) => cb(sc.fields[idx]));
		};
	}

	get el() {
		return this.container;
	}

	set updateSchema(cb: Update<MaybePersistedSchema>) {
		this._updateSchema = cb;
	}

	schemaChanged(schema: MaybePersistedSchema, errors: ValidateSchemaResult) {
		this.nameInput.value = schema.name;
		this.nameInput.error = errors.name?.[0];

		this.fields.fieldsChanged(schema.fields, errors);
	}
}

export class SchemaEditor {
	private editor = new Editor();

	get el() {
		return this.editor.el;
	}

	editSchema(schema: MaybePersistedSchema) {
		const schemaChanged = () => {
			const errors = validateSchema(schema);

			this.editor.schemaChanged(schema, errors);
		};

		const updateSchema = (cb: Callback<[typeof schema]>) => {
			cb(schema);
			console.debug({ schema });
			schemaChanged();
		};

		this.editor.updateSchema = updateSchema;
		schemaChanged();
	}
}
